/**
 * Computación Paralela
 *
 *
 * Práctica Stencil - Versión secuencial base
 */

// Muestra un display con la matrix (para matrices pequeñas)
// Compilar: make display
// #define SHOW_DISPLAY

// Bibliotecas estándard
#include <stdio.h>
#include <stdlib.h>
#include <math.h>

// Definiciones para permitir la compilación sin  
#ifdef _OPENMP
#include <omp.h>
#else
#define omp_get_thread_num() 0
#define omp_get_num_threads() 1
#define omp_set_num_threads(T) 0
#endif

// Biblioteca de funciones para computacion paralela
#include <cputils.h>

// Constantes de los pesos del stencil
#define WEIGHT_CENTER   (0.5   * (1.0/1.0) )
#define WEIGHT_ADJACENT (0.295 * (1.0/4.0) )
#define WEIGHT_CORNER   (0.195 * (1.0/4.0) )
#define MAX_RESID 0.001

/** Variable para el cálculo de residuos.  */
m_type resid;

/**
 * Genera la siguiente iteración
 * @param matrix_old El estado anterior.
 * @param matrix_new El nuevo estado.
 * @param rows El número de filas.
 * @param cols El número de columnas.
 */ 
void update(m_type * matrix_old, m_type * matrix_new, int rows, int cols);

/**
 * Suma los valores de matrix para verificar la solución
 * @param matrix La matrix
 * @param rows El número de filas.
 * @param cols El número de columnas.
 */
m_type check_sum(m_type * matrix, int rows, int cols);

/**
 * Función main
 */
int main(int argc, char *argv[]){

	//************************************************************
	// 0. Variables
	char * matrix_name; /**< Nombre de la matrix de entrada */
	int rows;           /**< Número de filas de la matriz */
	int cols;           /**< Número de columnas */
	m_type * matrix1;   /**< Matrix con los datos */
	m_type * matrix2;   /**< Copia de la matrix para calcular nuevos valores */
	double t_begin;     /**< Tiempo de inicio del calculo */
	double t_end;       /**< Tiempo de finalización */
	m_type sum;         /**< Suma de verificación */
	int iter = 0;		/**< Iteraciones realizadas hasta llegar a un punto estable */
	// Macros para acceder a las matrices
#define m1(i,j) (matrix1[(i)*(cols+2)+(j)])
#define m2(i,j) (matrix2[(i)*(cols+2)+(j)])
	//************************************************************
	
	//************************************************************
	// 1. Leer los parametros de entrada.
	if(argc != 2){
		fprintf(stderr,"USO: %s <matriz de datos>\n",argv[0]);
		exit(EXIT_FAILURE);
	}
	
	matrix_name = argv[1];
	//************************************************************

	
	//************************************************************
	// 2. Cargar la matriz
	// Matrix de tamaño rows+2 x cols+2 (halo de tamaño 1)
	// (Los datos están en [1,rows+1][1,cols+1])
	cp_read_matrix_size(matrix_name, &rows, &cols);
	printf(" Matriz %s: %dx%d\n",matrix_name, rows, cols);
	int matrix_size = (rows + 2) * (cols + 2);
	matrix1 = malloc(sizeof(m_type) * (size_t) matrix_size);
	cp_read_matrix(matrix_name, matrix1, 1);
	//************************************************************


	//************************************************************
	// 3. Copia de la matriz
	matrix2 = malloc(sizeof(m_type) * (size_t) matrix_size);
	//************************************************************

#ifdef SHOW_DISPLAY
	cp_display_create("Stencil", rows+2, cols+2);
	cp_display_draw_matrix(matrix1,CP_RGB(255,0,0),CP_RGB(0,255,0));
	cp_msleep(1000);
#endif

	//************************************************************
	// 4. Bucle principal
	t_begin = cp_Wtime(); 
	
	//int i,j;
	// 4.1 Nos mantenemos en el bucle mientras el residuo calculado 
	//  sea mayor que el residuo objetivo
	do {
		
		resid = 0.0;

		// Iteración del algoritmo
		update(matrix1,matrix2,rows,cols);
		
		// Actualizar la copia
	/*	#pragma omp parallel default(none)\
		shared(rows,cols,matrix1,matrix2)\
		private(j)
		{
		#pragma omp for
		for (i=1; i<rows+1; i++) {
				for (j=1; j<cols+1; j++) {
					m1(i,j) = m2 (i,j);
			}
		}
		}*/

		m_type * aux = matrix1;
		matrix1 = matrix2;
		matrix2 = aux;


#ifdef SHOW_DISPLAY
		cp_display_draw_matrix(matrix2,CP_RGB(255,0,0),CP_RGB(0,255,0));
		cp_msleep(50); // Para ver más despacio la evolucion del proceso
#endif

	iter++;

	} while (resid > MAX_RESID);
	//************************************************************
	
	
	//************************************************************
	// 5. Suma de verificación
	sum = check_sum(matrix1,rows,cols);
	t_end = cp_Wtime();
	//************************************************************
	
	
	//************************************************************
	// 6. Mostrar resultados
	printf(" Check sum: %f\n", sum);
	printf(" Iteraciones: %d\n", iter);
	printf(" Tiempo de ejecución: %f\n", t_end-t_begin);
	//************************************************************

#ifdef SHOW_DISPLAY	
	cp_msleep(250);
	cp_display_close();
#endif


	//************************************************************
	// 7. Liberar las matrices
	free(matrix1);
	free(matrix2);
	//************************************************************

#undef m1
#undef m2
	return EXIT_SUCCESS;
}



void update(m_type * matrix_old, m_type * matrix_new, int rows, int cols){

	//************************************************************
	// 1. Macros para acceder a las matrices
	#define m_old(i,j) (matrix_old[(i)*(cols+2)+(j)])
	#define m_new(i,j) (matrix_new[(i)*(cols+2)+(j)])
	//************************************************************

	//************************************************************
	// 2. Copia de los bordes
	int i, j;
	m_type tmpResid; //= fabs (m_old(i, j) - m_new(i, j));
	#pragma omp parallel default(none)\
	shared(rows,cols,matrix_new,matrix_old,resid)\
	private(j,tmpResid)
	{
	#pragma omp for //nowait
	for(i=1; i<rows+1; i++){
		m_old(i,0) = m_old(i,cols);
		m_old(i,cols+1) = m_old(i,1);
	}
	#pragma omp for //nowait
	for(j=0; j<cols+2; j++){
		m_old(0,j) = m_old(rows,j);
		m_old(rows+1,j) = m_old(1,j);
	}
/*
	#pragma omp sections
	//#pragma omp single
	{
	#pragma omp section
	m_old(0,0) = m_old(rows,0);
	#pragma omp section
	m_old(rows+1,0) = m_old(1,0);
	#pragma omp section
	m_old(0,cols+1) = m_old(rows,cols+1);
	#pragma omp section
	m_old(rows+1,cols+1) = m_old(1,cols+1);
	}*/
	
	//************************************************************

	//************************************************************
	// 3. Bucle principal
	#pragma omp for
	for(i=1; i<rows+1; i++){
		for(j=1; j<cols+1; j++){

			// 3.1 El nuevo valor es el anterior más la acción de los vecinos.
			m_new (i,j) = ( m_old (i, j) * WEIGHT_CENTER) +
						 
						((m_old (i-1, j) +
							m_old (i+1, j) +
							m_old (i, j-1) +
							m_old (i, j+1) ) * WEIGHT_ADJACENT) +
						 
						((m_old (i-1, j-1) +
							m_old (i-1, j+1) +
							m_old (i+1, j-1) +
							m_old (i+1, j+1) ) * WEIGHT_CORNER);
			
			// 3.2 Guardamos la mayor diferencia entre el valor antiguo y el actualizado
			{
			//m_type tmpResid = fabs (m_old(i, j) - m_new(i, j));
			tmpResid = fabs (m_old(i, j) - m_new(i, j));
			if ( tmpResid > resid ) resid = tmpResid;
			}
				
		}
	}
	}//pragma
	//************************************************************

	#undef m_old
	#undef m_new
}


m_type check_sum(m_type * matrix, int rows, int cols){

	//************************************************************
	// 1. Macros para acceder a la matriz
	#define m(i,j) (matrix[(i)*(cols+2)+(j)])
	//************************************************************

	//************************************************************
	// 2. Bucle con la suma
	m_type sum = 0;
	int i,j;
	for(i=1; i<rows+1; i++)
		for(j=1; j<cols+1; j++)
			sum += m(i,j);
	//************************************************************

	return sum;

	#undef m
}




